import logging
import sys
import itertools
import time
import os

import click
import click_log

from ..utils import bam_utils

import pysam


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("convert")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-o",
    "--output-bam",
    default="-",
    type=click.Path(exists=False),
    help="bam output  [default: stdout]",
)
@click.option(
    "-q",
    "--default-rq",
    type=float,
    default=-1,
    show_default=True,
    required=False,
    help="Default read quality for ingested reads (ranging from [-1,1])"
)
@click.option(
    "-g",
    "--read-group-id",
    type=str,
    help="Read group ID to set"
)
@click.option(
    "-s",
    "--sample-name",
    type=str,
    help="Sample name to set in the read group"
)
@click.option(
    "-l",
    "--min-length",
    type=int,
    default=0,
    required=False,
    help="Minimum length of reads to process"
)
@click.option(
    "-L",
    "--max-length",
    type=int,
    default=30000,
    required=False,
    help="Maximum length of reads to process"
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.argument("input-spec", required=True, type=click.Path(exists=True))
def main(output_bam, default_rq, read_group_id, sample_name, min_length, max_length, force, input_spec):
    """Convert reads from fastq{,.gz} files for use with `annotate`."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    # Check to see if the output files exist:
    bam_utils.check_for_preexisting_files(output_bam, exist_ok=force)

    # Get the list of input files
    files = _get_files(input_spec)

    # Create header for the output bam file:
    h = bam_utils.create_bam_header_with_program_group(
        logger.name,
        pysam.AlignmentHeader().from_dict({ 
            "HD": { "VN": "1.0" }, 
            "RG": [{ "ID": read_group_id, "SM": sample_name }],
        })
    )

    num_reads_seen, num_reads_written = 0, 0
    with pysam.AlignmentFile(output_bam, "wb", header=h) as bam_out:
        for file in files:
            with pysam.FastxFile(file) as fh:
                for entry in fh:
                    num_reads_seen += 1

                    if len(entry.sequence) >= min_length and len(entry.sequence) <= max_length:
                        r = pysam.AlignedSegment()
                        r.query_name = entry.name
                        r.query_sequence = entry.sequence
                        r.query_qualities = pysam.qualitystring_to_array(entry.quality)
                        r.flag = 4
                        r.mapping_quality = 255
                        r.set_tag("rq", default_rq)
                        r.set_tag("np", 1)
                        r.set_tag("RG", read_group_id)

                        bam_out.write(r)

                        num_reads_written += 1

    et = time.time()
    logger.info(f"Done. Elapsed time: {et - t_start:2.2f}s. "
                f"Reads seen: {num_reads_seen}. Reads written: {num_reads_written}.")


def _get_files(user_input):
    inputs = []

    if os.path.isdir(user_input):
        for root, dirs, files in os.walk(user_input):
            for file in files:
                if file.endswith(".fastq.gz") or file.endswith(".fq.gz"):
                    inputs.append(os.path.join(root, file))
    elif os.path.isfile(user_input):
        if user_input.endswith(".fastq.gz") or user_input.endswith(".fq.gz"):
            inputs.append(user_input)

    return inputs



import sys
import pysam
import filecmp

from pathlib import Path


def convert_sam_to_bam(sam_path, out_bam_path):
    with pysam.AlignmentFile(sam_path, "r", check_sq=False, require_index=False) as input_file:
        with pysam.AlignmentFile(out_bam_path, "wb", header=input_file.header) as out_bam_file:
            for r in input_file:
                out_bam_file.write(r)


def assert_read_tags_are_equal(actual_read, expected_read):

    actual_tag_val_dict = dict()
    actual_tag_type_dict = dict()
    for tag, val, tp in actual_read.get_tags(with_value_type=True):
        actual_tag_val_dict[tag] = val
        actual_tag_type_dict[tag] = tp

    expected_tag_val_dict = dict()
    expected_tag_type_dict = dict()
    for tag, val, tp in expected_read.get_tags(with_value_type=True):
        expected_tag_val_dict[tag] = val
        expected_tag_type_dict[tag] = tp

    assert len(actual_tag_val_dict) == len(expected_tag_val_dict), f"Read {actual_read.query_name}: Number of tags not equal: {len(actual_tag_val_dict)} != {len(expected_tag_val_dict)} || {actual_tag_val_dict} | {expected_tag_val_dict}"

    for tag in actual_tag_val_dict.keys():

        assert tag in expected_tag_val_dict, f"Read {actual_read.query_name}: Tag {tag} not in expected tags!"
        assert tag in actual_tag_type_dict, f"Read {actual_read.query_name}: Tag {tag} not in expected tags!"

        actual_val = actual_tag_val_dict[tag]
        actual_tp = actual_tag_type_dict[tag]

        expected_val = expected_tag_val_dict[tag]
        expected_tp = expected_tag_type_dict[tag]

        assert actual_val == expected_val, f"Read {actual_read.query_name}: Actual and expected tag values are not equal for tag {tag}: {actual_val} != {expected_val}"
        assert actual_tp == expected_tp, f"Read {actual_read.query_name}: Actual and expected tag types are not equal for tag {tag}: {actual_tp} != {expected_tp}"


def assert_reads_are_equal(actual_read, expected_read):

    # Go through most fields until tags:
    assert actual_read.query_name == expected_read.query_name, f"Read {actual_read.query_name}: Read names not equal: {actual_read.query_name} != {expected_read.query_name}"
    assert actual_read.flag == expected_read.flag, f"Read {actual_read.query_name}: Read flags not equal"
    assert actual_read.reference_name == expected_read.reference_name, f"Read {actual_read.query_name}: Contig names not equal:  {actual_read.reference_name} != {expected_read.reference_name}"
    assert actual_read.reference_start == expected_read.reference_start, f"Read {actual_read.query_name}: Start position not equal:  {actual_read.reference_start} != {expected_read.reference_start}"
    assert actual_read.mapping_quality == expected_read.mapping_quality, f"Read {actual_read.query_name}: Mapping qualities not equal:  {actual_read.mapping_quality} != {expected_read.mapping_quality}"
    assert actual_read.cigarstring == expected_read.cigarstring, f"Read {actual_read.query_name}: CIGAR strings not equal:  {actual_read.cigarstring} != {expected_read.cigarstring}"
    assert actual_read.next_reference_id == expected_read.next_reference_id, f"Read {actual_read.query_name}: RNEXT fields not equal:  {actual_read.next_reference_id} != {expected_read.next_reference_id}"
    assert actual_read.next_reference_start == expected_read.next_reference_start, f"Read {actual_read.query_name}: PNEXT fields not equal:  {actual_read.next_reference_start} != {expected_read.next_reference_start}"
    assert actual_read.template_length == expected_read.template_length, f"Read {actual_read.query_name}: TLEN fields not equal:  {actual_read.template_length} != {expected_read.template_length}"
    assert actual_read.query_sequence == expected_read.query_sequence, f"Read {actual_read.query_name}: Base sequences not equal"
    assert actual_read.query_qualities == expected_read.query_qualities, f"Read {actual_read.query_name}: Base qualities not equal"

    # Tags should be viewed as a collection, rather than ordered list.
    # As long as the values between the two reads are all present, we should call the tags equal:
    assert_read_tags_are_equal(actual_read, expected_read)


def assert_reads_files_equal(actual_file, expected_file, order_matters=False, compare_header=False):
    """Assert that the contents of the two given bam files are equivalent."""

    actual_file = Path(actual_file)
    expected_file = Path(expected_file)

    actual_file_flags = "rb" if actual_file.resolve().name.endswith(".bam") else "r"
    expected_file_flags = "rb" if expected_file.resolve().name.endswith(".bam") else "r"

    if order_matters and compare_header:
        assert filecmp.cmp(actual_file, expected_file)
        return

    with pysam.AlignmentFile(actual_file, actual_file_flags, check_sq=False, require_index=False) as actual_bam, \
        pysam.AlignmentFile(expected_file, expected_file_flags, check_sq=False, require_index=False) as expected_bam:

        if compare_header:
            assert actual_bam.header == expected_bam.header, "Headers are not equal."

        if order_matters:
            for read1, read2 in zip(actual_bam, expected_bam):
                assert_reads_are_equal(read1, read2)
        else:
            expected_reads = dict()

            for r in expected_bam:
                expected_reads[r.query_name] = r

            nreads_2 = 0
            for r in actual_bam:
                assert r.query_name in expected_reads, f"Read not in expected file: {r.query_name}"
                assert_reads_are_equal(r, expected_reads[r.query_name])
                nreads_2 += 1

            assert nreads_2 == len(expected_reads)


def assert_text_files_equal(actual_file, expected_file, header_prefix="#", order_matters=False, compare_header=False):
    """Assert that the contents of the two given bam files are equivalent."""

    line_no = 1

    if order_matters:
        if compare_header:
            assert filecmp.cmp(actual_file, expected_file)
            return
        else:
            with open(actual_file, 'r') as f_actual:
                with open(expected_file, 'r') as f_expected:


                    # Skip headers:
                    actual_line = f_actual.readline()
                    line_no += 1
                    expected_line = f_expected.readline()
                    while actual_line.startswith(header_prefix):
                        actual_line = f_actual.readline()
                        line_no += 1
                    while expected_line.startswith(header_prefix):
                        expected_line = f_expected.readline()

                    # Get the rest of the lines:

                    while actual_line and expected_line:
                        line_is_ok = actual_line == expected_line
                        if not line_is_ok:
                            print(f"ERROR: Line {line_no} not equal: Actual: {actual_line} || "
                                  f"Expected: {expected_line}", file=sys.stderr)

                        assert line_is_ok

                        actual_line = f_actual.readline()
                        expected_line = f_expected.readline()

                        line_no += 1

                    assert not actual_line and not expected_line
    else:
        if compare_header:
            print("ERROR: This type of comparison is invalid.")
            assert False
        else:
            with open(actual_file, 'r') as f_actual:
                with open(expected_file, 'r') as f_expected:

                    # Skip headers:
                    actual_line = f_actual.readline()
                    line_no += 1
                    expected_line = f_expected.readline()
                    while actual_line.startswith(header_prefix):
                        actual_line = f_actual.readline()
                        line_no += 1
                    while expected_line.startswith(header_prefix):
                        expected_line = f_expected.readline()

                    expected_lines = f_expected.readlines()

                    for i, line in enumerate(f_actual):
                        line_is_ok = line in expected_lines
                        if not line_is_ok:
                            print(f"ERROR: Line {line_no + i} found in expected file: Actual: {actual_line}",
                                  file=sys.stderr)
                        assert line_is_ok

                    has_lines_remaining = len(expected_lines) != i
                    if has_lines_remaining:
                        print(f"ERROR: number of lines is different in actual and expected files.",
                              file=sys.stderr)
                    assert not has_lines_remaining


def cat_file_to_pipe(filename, proc):
    with open(filename, "rb") as input_file:
        input_bytes = input_file.read()

        proc.stdin.write(input_bytes)
        proc.stdin.flush()

    proc.stdin.close()
    proc.wait()



import logging

import click
import click_log

import sys
import pkgutil

import longbow
from .meta import VERSION


logger = logging.getLogger("version")
click_log.basic_config(logger)
logger.handlers[0].formatter = logging.Formatter(
    "[%(levelname)s %(asctime)s %(name)8s] %(message)s", "%Y-%m-%d %H:%M:%S"
)


@click.group(name="longbow")
def main_entry():
    logger.info("Invoked via: longbow %s", " ".join(sys.argv))


@main_entry.command()
@click_log.simple_verbosity_option(logger)
def version():
    """Print the version of longbow."""
    click.echo(VERSION)


# Dynamically find and import sub-commands (allows for plugins at run-time):
# An alternative would be to iterate through the following:
# pkgutil.iter_modules([os.path.dirname((inspect.getmodule(sys.modules[__name__]).__file__))])
for p in [p for p in pkgutil.iter_modules(longbow.__path__) if p.ispkg]:
    exec(f"from .{p.name} import command as {p.name}")
    exec(f"main_entry.add_command({p.name}.main)")


if __name__ == "__main__":
    main_entry()  # pylint: disable=E1120



import pytest

import pysam
import tempfile
import os
import json

from longbow.utils import bam_utils
from longbow.utils import model


TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
)


@pytest.mark.parametrize("model_name, expected_value", [
    ["mas_15_sc_10x5p_single_none", True],
    ["mas_15_sc_10x3p_single_none", True],
    ["mas_15_bulk_10x5p_single_internal", False],
    ["mas_10_sc_10x5p_single_none", True],
    ["mas_15_spatial_slide-seq_single_none", False],
    ["mas_15_bulk_teloprimeV2_single_none", False],
    ["isoseq_1_sc_10x5p_single_none", True],
])
def test_has_cell_barcode_annotation(model_name, expected_value):
    lb_model = model.LibraryModel.build_pre_configured_model(model_name)
    assert lb_model.has_cell_barcode_annotation == expected_value


@pytest.mark.parametrize("model_name, expected_value", [
    ["mas_15_sc_10x5p_single_none", True],
    ["mas_15_sc_10x3p_single_none", True],
    ["mas_15_bulk_10x5p_single_internal", True],
    ["mas_10_sc_10x5p_single_none", True],
    ["mas_15_spatial_slide-seq_single_none", True],
    ["mas_15_bulk_teloprimeV2_single_none", False],
    ["isoseq_1_sc_10x5p_single_none", True],
])
def test_has_umi_annotation(model_name, expected_value):
    lb_model = model.LibraryModel.build_pre_configured_model(model_name)
    assert lb_model.has_umi_annotation == expected_value



import logging
import sys
import itertools

import time

import click
import click_log
import tqdm

import pysam
import multiprocessing as mp

from inspect import getframeinfo, currentframe, getdoc

import longbow.utils.constants
from ..utils import bam_utils
from ..utils import model as LongbowModel
from ..utils.model import LibraryModel

from ..annotate.command import get_segments

from ..meta import VERSION

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("scsplit")
click_log.basic_config(logger)


__DEFAULT_DUMMY_CBC = "CTGCCTAACCTGATCC"
__DEFAULT_OUT_BASE_NAME = logger.name
__DEFAULT_UMI_LENGTH = 10


__OUT_READ_FILE_SUFFIX = "_mates"
__OUT_WHITELIST_FILE_SUFFIX = "_whitelist.txt"


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=mp.cpu_count() - 1,
    show_default=True,
    help="number of threads to use (0 for all)",
)
@click.option(
    "-o",
    "--output-base-name",
    default=__DEFAULT_OUT_BASE_NAME,
    type=str,
    help=f"base name for output files [default: {__DEFAULT_OUT_BASE_NAME}]",
)
@click.option(
    "-c",
    "--cell-barcode",
    default=__DEFAULT_DUMMY_CBC,
    type=str,
    help=f"dummy cell barcode to use for the dataset [default: {__DEFAULT_DUMMY_CBC}, "
         f"length: {len(__DEFAULT_DUMMY_CBC)}]",
)
@click.option(
    "-u",
    "--umi-length",
    default=__DEFAULT_UMI_LENGTH,
    type=int,
    show_default=True,
    help=f"length of the UMI from this library",
)
@click.option(
    "-b",
    "--write-bam",
    is_flag=True,
    default=False,
    show_default=True,
    help=f"Write out an annotated bam file in addition to the mates files.",
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force scsplit to run on the input bam without checking for compatibility."
)
@click.option(
    "-m",
    "--model",
    default=longbow.utils.constants.DEFAULT_MODEL,
    show_default=True,
    help="The model to use for annotation.  If the given value is a pre-configured model name, then that "
         "model will be used.  Otherwise, the given value will be treated as a file name and Longbow will attempt to "
         "read in the file and create a LibraryModel from it.  Longbow will assume the contents are the configuration "
         "of a LibraryModel as per LibraryModel.to_json()."
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(threads, output_base_name, cell_barcode, umi_length, force, model, write_bam, input_bam):
    """Create files for use in `alevin` for single-cell analysis.
    This tool coerces a set of reads from a single source into a format that `alevin` can ingest.
    
    Segment names are assumed to be those in the default model (utils/model.py).
    
    INPUT_BAM should contain reads that have been processed by `longbow segment`.

    The output from this tool consists of several files:
        OUTPUT_BASE_NAME_mates_1.fastq:
            A file containing partial sequences for all reads in the given input file.  These partial reads consist of the 
            dummy cell barcode + detected UMI for each read in the given input file.
        OUTPUT_BASE_NAME_mates_2.fastq:
            A file containing partial sequences for all reads in the given input file.  These partial reads consist of the 
            transcript sequences for all reads in the given input file.  Transcript sequences include data after the UMI 
            and before the Poly-A tail.  All bases outside of this range are excluded from the output.
        OUTPUT_BASE_NAME_whitelist.txt:
            A whitelist file for alevin containing the given dummy cell barcode. 
    """

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    threads = mp.cpu_count() if threads <= 0 or threads > mp.cpu_count() else threads
    logger.info(f"Running with {threads} worker subprocess(es)")

    # Get our model:
    if LibraryModel.has_prebuilt_model(model):
        lb_model = LibraryModel.build_pre_configured_model(model)
    else:
        logger.info(f"Loading model from json file: %s", model)
        lb_model = LibraryModel.from_json_file(model)
    logger.info(f"Using %s: %s", model, lb_model.description)
    model = lb_model

    # Configure process manager:
    # NOTE: We're using processes to overcome the Global Interpreter Lock.
    manager = mp.Manager()
    process_input_data_queue = manager.Queue(threads)
    results = manager.Queue()

    # Start worker sub-processes:
    worker_process_pool = []
    for _ in range(threads):
        p = mp.Process(
            target=_sub_process_work_fn, args=(process_input_data_queue, results, umi_length, model, write_bam)
        )
        p.start()
        worker_process_pool.append(p)

    pysam.set_verbosity(0)  # silence message about the .bai file not being found
    with pysam.AlignmentFile(
        input_bam, "rb", check_sq=False, require_index=False
    ) as bam_file, tqdm.tqdm(
        desc="Progress",
        unit=" read",
        colour="green",
        file=sys.stderr,
        leave=False,
        disable=not sys.stdin.isatty(),
    ) as pbar:
        if force:
            logger.info("Force mode - skipping bam header check for compatibility")
        else:
            # Make sure we're given an input bam file we can work with:
            if not _validate_input_bam(bam_file.header):
                # Bad news - we have to quit.
                # let's try to do it nicely:
                for r in (None,) * threads:
                    process_input_data_queue.put(r)

                # Wait for our input jobs to finish:
                for p in worker_process_pool:
                    p.join()
                sys.exit(1)

        # Get our model:
        if model is None:
            lb_model = LibraryModel.from_json_obj(bam_utils.get_model_from_bam_header(bam_file.header))
        elif model is not None and LibraryModel.has_prebuilt_model(model):
            lb_model = LibraryModel.build_pre_configured_model(model)
        else:
            lb_model = LibraryModel.from_json_file(model)

        logger.info(f"Using %s: %s", lb_model.name, lb_model.description)

        # Get our header from the input bam file:
        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header, models=[model])

        # Start output worker:
        res = manager.dict({"num_reads_processed": 0})
        output_worker = mp.Process(
            target=_sub_process_write_fn,
            args=(
                results,
                output_base_name,
                cell_barcode,
                pbar,
                res,
                write_bam,
                out_header
            ),
        )
        output_worker.start()

        # Add in a sentinel value at the end of the queue - one for each subprocess - so we guarantee
        # that all subprocesses will exit:
        iter_data = itertools.chain(bam_file, (None,) * threads)
        for r in iter_data:
            if r is not None:
                process_input_data_queue.put(r.to_string())
            else:
                process_input_data_queue.put(r)

        # Wait for our input jobs to finish:
        for p in worker_process_pool:
            p.join()

        # Now that our input processes are done, we can add our exit sentinel onto the output queue and
        # wait for that process to end:
        results.put(None)
        output_worker.join()

    # Write out our CBC whitelist file:
    with open(f"{output_base_name}{__OUT_WHITELIST_FILE_SUFFIX}", "w") as f:
        f.write(f"{cell_barcode}\n")

    logger.info(f"Processed {res['num_reads_processed']} reads.")
    logger.info(f"CBC length: {len(cell_barcode)}.")
    logger.info(f"UMI length: {umi_length}.")
    logger.info(f"Done. Elapsed time: %2.2fs.", time.time() - t_start)


def _validate_input_bam(input_bam_header):
    """Check that the given input_bam_header contains an `longbow segment` program group."""
    in_bam_header_dict = input_bam_header.to_dict()
    if "PG" not in in_bam_header_dict:
        logger.warning("Could not find PG entry in header.  Cannot confirm that this file is compatible.")
    else:
        found_segment_cmd = False
        for info in [item for item in in_bam_header_dict["PG"]]:
            if "PN" not in info:
                continue
            if info["PN"] == "longbow" and info["ID"].split("-")[1] == "segment":
                found_segment_cmd = True
                break
        if not found_segment_cmd:
            logger.error(
                "Input bam file header does not indicate that it was created by longbow segment.  "
                "This tool requires `longbow segment` reads as input data.")
            return False
    return True


def _get_start_segment_from_list(seg_list, model, read_name):
    """Get the start segment segment from the list of SegmentInfo objects based on the given model.
    If no start segment is found, returns None."""

    # The start segment should be the first matching segment:
    for s in seg_list:
        if s.name in model.start_element_names:
            return s

    logger.warning("Could not process read: %s - No start segment found (start names: %s).",
                   read_name, model.start_element_names)
    return None


def _get_end_segment_from_list(seg_list, model, read_name):
    """Get the end segment segment from the list of SegmentInfo objects based on the given model.
    If no start segment is found, returns None."""

    # The end segment should be the last matching segment, so we
    # iterate from the end to the start of the list:
    for s in reversed(seg_list):
        if s.name in model.end_element_names:
            return s

    logger.warning("Could not process read: %s - No end segment found (end names: %s).",
                   read_name, model.start_element_names)
    return None


def _sub_process_work_fn(in_queue, out_queue, umi_length, array_model, do_bam_out):
    """Function to run in each subprocess.
    Extracts and returns all segments from an input read."""
    while True:
        # Wait until we get some data.
        # Note: Because we have a sentinel value None inserted at the end of the input data for each
        #       subprocess, we don't have to add a timeout - we're guaranteed each process will always have
        #       at least one element.
        raw_data = in_queue.get()

        # Check for exit sentinel:
        if raw_data is None:
            return

        # Unpack our data here:
        read = pysam.AlignedSegment.fromstring(
            raw_data, pysam.AlignmentHeader.from_dict(dict())
        )
        _, segments = get_segments(read)

        # Get start element position
        # (for MAS-seq it's the 10x adapter)
        start_segment = _get_start_segment_from_list(segments, array_model, read.query_name)
        if start_segment is None:
            continue

        # Get the end element position:
        # (for MAS-seq it's the Poly-a)
        end_segment = _get_end_segment_from_list(segments, array_model, read.query_name)
        if end_segment is None:
            continue

        # Now we grab the bases just after the 10x adapter as the UMI
        # and the bases between the UMI and the poly A for the transcript

        # Note: Positions are inclusive so we must add 1 to the end position to get that base as well:
        umi_start = start_segment.end+1
        umi_end = umi_start + umi_length
        umi_bases = read.query_sequence[umi_start:umi_end]
        umi_quals = "".join([chr(i + 33) for i in read.query_alignment_qualities[umi_start:umi_end]])

        transcript_bases = read.query_sequence[umi_end:end_segment.start]
        transcript_quals = "".join(
            [chr(i + 33) for i in read.query_alignment_qualities[umi_end:end_segment.start]]
        )

        # Place our data on the output queue:
        if do_bam_out:
            out_queue.put(
                tuple([read.query_name, umi_bases, umi_quals, transcript_bases, transcript_quals, read.to_string()])
            )
        else:
            out_queue.put(
                tuple([read.query_name, umi_bases, umi_quals, transcript_bases, transcript_quals])
            )


def _sub_process_write_fn(
    out_queue,
    out_base_name,
    cell_barcode,
    pbar,
    res,
    do_bam_out,
    out_bam_header
):
    """Thread / process fn to write out all our data."""

    try:
        if do_bam_out:
            out_bam_file = pysam.AlignmentFile(f"{out_base_name}.cbc_umi_annotated.bam", "wb", header=out_bam_header)
        with open(f"{out_base_name}{__OUT_READ_FILE_SUFFIX}1.fastq", "w") as mates1_file, \
             open(f"{out_base_name}{__OUT_READ_FILE_SUFFIX}2.fastq", "w") as mates2_file:

            while True:
                # Wait for some output data:
                raw_data = out_queue.get()

                # Check for exit sentinel:
                if raw_data is None:
                    break

                # Unpack data:
                if do_bam_out:
                    read_name, umi_bases, umi_quals, transcript_bases, transcript_quals, read_string = raw_data
                else:
                    read_name, umi_bases, umi_quals, transcript_bases, transcript_quals = raw_data

                # Create mates1 and mates2 records:
                mates_1_record = pysam.FastxRecord(
                    name=read_name,
                    sequence=cell_barcode + umi_bases,
                    quality=(chr(33 + 60) * len(cell_barcode)) + umi_quals
                )
                mates_2_record = pysam.FastxRecord(
                    name=read_name,
                    sequence=transcript_bases,
                    quality=transcript_quals
                )

                # Write out mates1 and mates2 records:
                mates1_file.write(str(mates_1_record))
                mates1_file.write("\n")
                mates2_file.write(str(mates_2_record))
                mates2_file.write("\n")

                if do_bam_out:
                    read = pysam.AlignedSegment.fromstring(
                        read_string, pysam.AlignmentHeader.from_dict(dict())
                    )

                    read.set_tag("CR", cell_barcode)
                    read.set_tag("UR", umi_bases)

                    out_bam_file.write(read)

                # Increment our counters:
                res["num_reads_processed"] += 1
                pbar.update(1)

                # Obligatory log message:
                logger.debug("Processed read: %s", read_name)
    finally:
        if do_bam_out:
            out_bam_file.close()



import pytest
import os
import sys
import subprocess
import tempfile

from click.testing import CliRunner

from longbow.__main__ import main_entry as longbow

from ..utils import cat_file_to_pipe

TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
) + os.path.sep


@pytest.fixture(scope="module", params=[
    (TEST_DATA_FOLDER + "mas15_test_input.bam", "mas_15_sc_10x5p_single_none"),
    (TEST_DATA_FOLDER + "mas10_test_input.bam", "mas_10_sc_10x5p_single_none"),
])
def annotated_bam_file_from_pipeline(request):
    input_bam, model_name = request.param

    with tempfile.NamedTemporaryFile(delete=True) as annotate_bam:

        runner = CliRunner()

        result_annotate = runner.invoke(longbow, ["annotate", "-m", model_name, "-f", "-o", annotate_bam.name, input_bam])
        assert result_annotate.exit_code == 0

        # Yield file here so that when we return, we get to clean up automatically
        yield annotate_bam.name


def test_demultiplex_from_file(tmpdir, annotated_bam_file_from_pipeline):
    args = ["demultiplex", "-d", "YN", "-o", "demux", annotated_bam_file_from_pipeline]

    runner = CliRunner()
    result = runner.invoke(longbow, args)

    assert result.exit_code == 0


@pytest.mark.skip(reason="this test is broken and I don't know why")
def test_demultiplex_from_pipe(tmpdir, annotated_bam_file_from_pipeline):
    proc = subprocess.Popen(
        [ sys.executable, "-m", "longbow", "demultiplex", "-d", "YN", "-o", "demux", annotated_bam_file_from_pipeline ],
        stdin=subprocess.PIPE
    )

    cat_file_to_pipe(annotated_bam_file_from_pipeline, proc)

    assert proc.returncode == 0



import logging
import queue
import time
import sys
import os
import itertools
import tempfile
import enum

import click
import click_log

import pysam
import multiprocessing as mp

from collections import defaultdict

from tqdm import tqdm

from construct import *

import longbow.utils.constants
from ..utils import bam_utils, barcode_utils
from ..utils.model import LibraryModel


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("correct")
click_log.basic_config(logger)


class BarcodeResolutionFailure(enum.Enum):
    NO_RAW_BARCODE = enum.auto()
    AMBIGUOUS = enum.auto()
    NO_MATCH_IN_LEV_DIST = enum.auto()


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(),
    help="BAM .pbi index file",
)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=mp.cpu_count() - 1,
    show_default=True,
    help="number of threads to use (0 for all)",
)
@click.option(
    "-o",
    "--output-bam",
    default="-",
    type=click.Path(exists=False),
    help="annotated bam output  [default: stdout]",
)
@click.option(
    "-m",
    "--model",
    help="The model to use for annotation.  If not specified, it will be autodetected from "
         "the BAM header.  If the given value is a pre-configured model name, then that "
         "model will be used.  Otherwise, the given value will be treated as a file name "
         "and Longbow will attempt to read in the file and create a LibraryModel from it.  "
         "Longbow will assume the contents are the configuration of a LibraryModel as per "
         "LibraryModel.to_json()."
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.option(
    '-r',
    '--restrict-to-allowlist',
    is_flag=True,
    default=True,
    show_default=True,
    help="Restrict barcode correction possibilities to only those on the allowlist."
)
@click.option(
    "-b",
    "--barcode-tag",
    type=str,
    default=longbow.utils.constants.READ_BARCODE_TAG,
    show_default=True,
    help="The tag from which to read the uncorrected barcode."
)
@click.option(
    "-c",
    "--corrected-tag",
    type=str,
    default=longbow.utils.constants.READ_BARCODE_CORRECTED_TAG,
    show_default=True,
    help="The tag in which to store the corrected barcode."
)
@click.option(
    "-a",
    "--allow-list",
    required=True,
    type=click.Path(exists=True),
    help="List of allowed barcodes for specified tag (.txt, .txt.gz).",
)
@click.option(
    "--barcode-freqs",
    type=click.Path(exists=True),
    required=False,
    help="TSV file containing barcodes and the frequencies associated with them in the data (BARCODE\tFREQ).  "
         "If not provided, barcode freqs will be uniformly seeded by the barcode whitelist.  "
         "NOTE: If barcodes in this freqs file are not in the allow list and the `-r` flag is not given, it is possible"
         "to end up with reads that have barcodes which where corrected to values that are not on the allow list.",
)
@click.option(
    "--max-hifi-dist",
    type=int,
    default=2,
    show_default=True,
    help="Maximum levenshtein distance to allow for hifi/CCS reads during correction.",
)
@click.option(
    "--max-clr-dist",
    type=int,
    default=3,
    show_default=True,
    help="Maximum levenshtein distance to allow for CLR (ccs uncorrected) reads during correction.",
)
@click.option(
    "--ccs-corrected-rq-threshold",
    type=float,
    default=0,
    show_default=True,
    help="Value of the `rq` tag above which reads are considered to be CCS corrected (Hifi).",
)
@click.option(
    "--barcode-uncorrectable-bam",
    default="/dev/null",
    show_default=True,
    type=click.Path(exists=False),
    help="File to which to write all reads with barcodes that could not be corrected.",
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(pbi, threads, output_bam, model, force, restrict_to_allowlist, barcode_tag, corrected_tag, allow_list,
         barcode_freqs, max_hifi_dist, max_clr_dist, ccs_corrected_rq_threshold, barcode_uncorrectable_bam, input_bam):
    """Correct tag to values provided in barcode allowlist."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    # Check to see if the output files exist:
    bam_utils.check_for_preexisting_files(output_bam, exist_ok=force)

    # Decide if we're disabling progress bars:
    disable_pbar = not sys.stdin.isatty()

    logger.info(f"Writing reads with corrected barcodes to: {output_bam}")
    logger.info(f"Writing reads with barcodes that could not be corrected to: {barcode_uncorrectable_bam}")

    threads = mp.cpu_count() if threads <= 0 or threads > mp.cpu_count() else threads
    logger.info(f"Running with {threads} worker subprocess(es)")

    logger.info(f"Using CCS Levenshtein distance threshold: {max_hifi_dist}")
    logger.info(f"Using CLR Levenshtein distance threshold: {max_clr_dist}")

    # Load barcode allow list:
    barcode_allow_list = barcode_utils.load_barcode_allowlist(allow_list, disable_pbar=disable_pbar)

    # Load number of reads, if pbi exists:
    pbi = f"{input_bam.name}.pbi" if pbi is None else pbi
    num_reads = bam_utils.load_read_count(pbi) if os.path.exists(pbi) else None

    # Create queues for data:
    # NOTE: We're using processes to overcome the Global Interpreter Lock.
    queue_size = threads * 2 if threads < 10 else 20
    manager = mp.Manager()
    process_input_data_queue = manager.Queue(maxsize=queue_size)
    # process_input_data_queue = manager.Queue()
    results = manager.Queue()

    pysam.set_verbosity(0)  # silence message about the .bai file not being found
    with pysam.AlignmentFile(input_bam, "rb", check_sq=False, require_index=False) as bam_file:

        # Load our model:
        if model is None:
            lb_model = LibraryModel.from_json_obj(bam_utils.get_model_from_bam_header(bam_file.header))
        elif model is not None and LibraryModel.has_prebuilt_model(model):
            lb_model = LibraryModel.build_pre_configured_model(model)
        else:
            lb_model = LibraryModel.from_json_file(model)
        logger.info(f"Using %s: %s", lb_model.name, lb_model.description)

        # Get our barcode length:
        barcode_length = _get_barcode_tag_length_from_model(lb_model, barcode_tag)

        # Generate our symspellpy index:
        if not barcode_freqs:
            logger.info("No barcode freq file provided.")
            barcode_freqs = _generate_barcode_freqs_from_allow_list(barcode_allow_list)
            logger.info(f"Generated uniform frequencies from allow list in file: {barcode_freqs}")
        else:
            logger.info(f"Using barcode frequencies from: {barcode_freqs}")
            if restrict_to_allowlist:
                logger.info("Filtering barcode frequencies file to only those on the allow list...")
                st = time.time()
                barcode_freqs = _filter_barcode_freqs_by_allow_list(barcode_freqs, barcode_allow_list)
                logger.info(f"Barcode freqs filtered in {time.time() - st:2.4f}s")
            else:
                logger.warning("Allowing ANY barcode in freqs file to seed corrections.  "
                               "THIS MAY PRODUCE NON-ALLOWLIST BARCODES IN RESULTING CORRECTED DATA!")

        # Start worker sub-processes:
        worker_process_pool = []
        for _ in range(threads):
            p = mp.Process(
                target=_correct_barcode_fn,
                args=(process_input_data_queue, results, bam_file.header.to_dict(), barcode_tag, corrected_tag,
                      barcode_length, barcode_allow_list, barcode_freqs, max_hifi_dist, max_clr_dist,
                      ccs_corrected_rq_threshold)
            )
            p.start()
            worker_process_pool.append(p)

        # Get header for output file:
        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header, models=[lb_model])

        # Start output worker:
        res = manager.dict({"num_ccs_reads": 0, "num_ccs_reads_corrected": 0, "num_ccs_reads_raw_was_correct": 0,
                            "num_clr_reads": 0, "num_clr_reads_corrected": 0, "num_clr_reads_raw_was_correct": 0,
                            "num_ccs_with_barcodes": 0, "num_clr_with_barcodes": 0,
                            "num_ccs_could_not_correct_ambiguous": 0, "num_ccs_could_not_correct_no_match": 0,
                            "num_clr_could_not_correct_ambiguous": 0, "num_clr_could_not_correct_no_match": 0, })
        output_worker = mp.Process(
            target=_write_thread_fn,
            args=(results, out_header, output_bam, barcode_uncorrectable_bam, barcode_tag, corrected_tag,
                  ccs_corrected_rq_threshold, num_reads, disable_pbar, res),
        )
        output_worker.start()

        # Add in a sentinel value at the end of the queue - one for each subprocess - so we guarantee
        # that all subprocesses will exit:
        iter_data = itertools.chain(bam_file, (None,) * threads)
        for r in iter_data:
            if r is not None:
                process_input_data_queue.put(r.to_string())
            else:
                process_input_data_queue.put(r)

        # Wait for our input jobs to finish:
        for p in worker_process_pool:
            p.join()

        # Now that our input processes are done, we can add our exit sentinel onto the output queue and
        # wait for that process to end:
        results.put(None)
        output_worker.join()

    # Print out some stats:

    stat_prefix = "STATS: "

    total_reads = res["num_ccs_reads"] + res["num_clr_reads"]
    logger.info(f"{stat_prefix}Total reads seen: {total_reads}")

    total_with_barcodes = res["num_ccs_with_barcodes"] + res["num_clr_with_barcodes"]
    count_str, pct_str = _get_field_count_and_percent_string(total_with_barcodes, total_reads)
    logger.info(f"{stat_prefix}Reads with {barcode_tag} barcodes: {count_str} {pct_str}")

    total_without_barcodes = total_reads - total_with_barcodes
    count_str, pct_str = _get_field_count_and_percent_string(total_without_barcodes, total_reads)
    logger.info(f"{stat_prefix}Reads without {barcode_tag} barcodes: {count_str} {pct_str}")

    total_corrected = res["num_ccs_reads_corrected"] + res["num_clr_reads_corrected"]
    count_str, pct_str = _get_field_count_and_percent_string(total_corrected, total_reads)
    logger.info(f"{stat_prefix}Reads able to be corrected into {corrected_tag} tag: {count_str} {pct_str}")

    total_uncorrected = total_reads - (res["num_ccs_reads_corrected"] + res["num_clr_reads_corrected"])
    count_str, pct_str = _get_field_count_and_percent_string(total_uncorrected, total_reads)
    logger.info(f"{stat_prefix}Reads unable to be corrected into {corrected_tag} tag: {count_str} {pct_str}")

    total_uncorrected_ambiguous = res["num_ccs_could_not_correct_ambiguous"] + res["num_clr_could_not_correct_ambiguous"]
    count_str, pct_str = _get_field_count_and_percent_string(total_uncorrected_ambiguous, total_reads)
    count_str2, pct_str2 = _get_field_count_and_percent_string(total_uncorrected_ambiguous, total_uncorrected)
    logger.info(f"{stat_prefix}Reads unable to be corrected into {corrected_tag} tag - Ambiguous (of total reads): {count_str} {pct_str}")
    logger.info(f"{stat_prefix}Reads unable to be corrected into {corrected_tag} tag - Ambiguous (of uncorrected reads): {count_str2} {pct_str2}")

    total_uncorrected_no_match = res["num_ccs_could_not_correct_no_match"] + res["num_clr_could_not_correct_no_match"]
    count_str, pct_str = _get_field_count_and_percent_string(total_uncorrected_no_match, total_reads)
    count_str2, pct_str2 = _get_field_count_and_percent_string(total_uncorrected_no_match, total_uncorrected)
    logger.info(f"{stat_prefix}Reads unable to be corrected into {corrected_tag} tag - No Match in Lev Dist (of total reads): {count_str} {pct_str}")
    logger.info(f"{stat_prefix}Reads unable to be corrected into {corrected_tag} tag - No Match in Lev Dist (of uncorrected reads): {count_str2} {pct_str2}")

    total_were_already_correct = res["num_ccs_reads_raw_was_correct"] + res["num_clr_reads_raw_was_correct"]
    count_str, pct_str = _get_field_count_and_percent_string(total_were_already_correct, total_reads)
    logger.info(f"{stat_prefix}Reads with already correct {barcode_tag} barcodes: {count_str} {pct_str}")

    for read_type in ['ccs', 'clr']:

        logger.info("=" * 80)

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_reads"], total_reads)
        logger.info(f"{stat_prefix}{read_type.upper()} reads seen: {count_str} {pct_str}")

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_with_barcodes"], total_reads)
        count_str2, pct_str2 = _get_field_count_and_percent_string(res[f"num_{read_type}_with_barcodes"], res[f"num_{read_type}_reads"])
        logger.info(f"{stat_prefix}{read_type.upper()} reads with {barcode_tag} barcodes (of total reads): {count_str} {pct_str}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads with {barcode_tag} barcodes (of {read_type} reads): {count_str2} {pct_str2}")

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_reads"] - res[f"num_{read_type}_with_barcodes"], total_reads)
        count_str2, pct_str2 = _get_field_count_and_percent_string(res[f"num_{read_type}_reads"] - res[f"num_{read_type}_with_barcodes"], res[f"num_{read_type}_reads"])
        logger.info(f"{stat_prefix}{read_type.upper()} reads without {barcode_tag} barcodes (of total reads): {count_str} {pct_str}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads without {barcode_tag} barcodes (of {read_type} reads): {count_str2} {pct_str2}")

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_reads_corrected"], total_reads)
        count_str2, pct_str2 = _get_field_count_and_percent_string(res[f"num_{read_type}_reads_corrected"], res[f"num_{read_type}_reads"])
        logger.info(f"{stat_prefix}{read_type.upper()} reads able to be corrected into {corrected_tag} tag (of total reads): {count_str} {pct_str}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads able to be corrected into {corrected_tag} tag (of {read_type} reads): {count_str2} {pct_str2}")

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_reads"] - res[f"num_{read_type}_reads_corrected"], total_reads)
        count_str2, pct_str2 = _get_field_count_and_percent_string(res[f"num_{read_type}_reads"] - res[f"num_{read_type}_reads_corrected"], res[f"num_{read_type}_reads"])
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag (of total reads): {count_str} {pct_str}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag (of {read_type} reads): {count_str2} {pct_str2}")

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_could_not_correct_ambiguous"], total_reads)
        count_str2, pct_str2 = _get_field_count_and_percent_string(res[f"num_{read_type}_could_not_correct_ambiguous"], res[f"num_{read_type}_reads"])
        count_str3, pct_str3 = _get_field_count_and_percent_string(res[f"num_{read_type}_could_not_correct_ambiguous"], res[f"num_{read_type}_reads"] - res[f"num_{read_type}_reads_corrected"])
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag - Ambiguous (of total reads): {count_str} {pct_str}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag - Ambiguous (of {read_type} reads): {count_str2} {pct_str2}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag - Ambiguous (of {read_type} uncorrected reads): {count_str3} {pct_str3}")

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_could_not_correct_no_match"], total_reads)
        count_str2, pct_str2 = _get_field_count_and_percent_string(res[f"num_{read_type}_could_not_correct_no_match"], res[f"num_{read_type}_reads"])
        count_str3, pct_str3 = _get_field_count_and_percent_string(res[f"num_{read_type}_could_not_correct_no_match"], res[f"num_{read_type}_reads"] - res[f"num_{read_type}_reads_corrected"])
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag - No Match in Lev Dist (of total reads): {count_str} {pct_str}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag - No Match in Lev Dist (of {read_type} reads): {count_str2} {pct_str2}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads unable to be corrected into {corrected_tag} tag - No Match in Lev Dist (of {read_type} uncorrected reads): {count_str3} {pct_str3}")

        count_str, pct_str = _get_field_count_and_percent_string(res[f"num_{read_type}_reads_raw_was_correct"], total_reads)
        count_str2, pct_str2 = _get_field_count_and_percent_string(res[f"num_{read_type}_reads_raw_was_correct"], res[f"num_{read_type}_reads"])
        logger.info(f"{stat_prefix}{read_type.upper()} reads with already correct {corrected_tag} tag (of total reads): {count_str} {pct_str}")
        logger.info(f"{stat_prefix}{read_type.upper()} reads with already correct {corrected_tag} tag (of {read_type} reads): {count_str2} {pct_str2}")

    et = time.time()
    logger.info(f"Done. Elapsed time: {et - t_start:2.2f}s. "
                f"Overall processing rate: {total_reads/(et - t_start):2.2f} reads/s.")


def _get_field_count_and_percent_string(count, total):
    count_str = f"{count}/{total}"
    pct_str = f"({0:2.4f}%)" if total == 0 else f"({100.0*count/total:2.4f}%)"

    return count_str, pct_str


def _write_thread_fn(data_queue, out_bam_header, out_bam_file_name, barcode_uncorrectable_bam,
                     barcode_tag, corrected_tag, ccs_corrected_rq_threshold, num_reads, disable_pbar, res):
    """Thread / process fn to write out all our data."""

    with pysam.AlignmentFile(out_bam_file_name, "wb", header=out_bam_header) as out_bam_file, \
        pysam.AlignmentFile(barcode_uncorrectable_bam, "wb", header=out_bam_header) as barcode_uncorrectable_bam, \
        tqdm(
            desc="Progress",
            unit=" read",
            colour="green",
            file=sys.stderr,
            disable=disable_pbar,
            total=num_reads,
            leave=False
        ) as pbar:

        while True:
            # Wait for some output data:
            raw_data = data_queue.get()

            # Check for exit sentinel:
            if raw_data is None:
                break
            # Should really never be None, but just in case:
            elif raw_data is None:
                continue

            # Unpack data:
            read, num_segments, num_corrected_segments, has_barcode = raw_data
            read = pysam.AlignedSegment.fromstring(read, out_bam_header)

            # Save ourselves some trouble later:
            ccs_type_stat_string = "ccs" if read.get_tag("rq") > ccs_corrected_rq_threshold else "clr"

            # Write our our read:
            if read.get_tag(longbow.utils.constants.COULD_CORRECT_BARCODE_TAG):
                out_bam_file.write(read)
            else:
                # "num_ccs_could_not_correct_ambiguous": 0, "num_ccs_could_not_correct_no_match"
                # Determine what the problem in correction was for our stats:
                correction_failure = read.get_tag(corrected_tag)

                # NOTE: using the enum.value field here because of the TYPE of the barcode tag.
                if correction_failure == BarcodeResolutionFailure.AMBIGUOUS.value:
                    res[f"num_{ccs_type_stat_string}_could_not_correct_ambiguous"] += 1
                elif correction_failure == BarcodeResolutionFailure.NO_MATCH_IN_LEV_DIST.value:
                    res[f"num_{ccs_type_stat_string}_could_not_correct_no_match"] += 1
                elif correction_failure == BarcodeResolutionFailure.NO_RAW_BARCODE.value:
                    # This case is already accounted for.
                    pass
                else:
                    raise RuntimeError(f"Unknown BarcodeResolutionFailure type ({correction_failure})!  "
                                       f"This should never happen!")

                # Set our corrected barcode tag to the unlabeled placeholder:
                read.set_tag(corrected_tag, longbow.utils.constants.UNLABELED_BARCODE)
                barcode_uncorrectable_bam.write(read)

            # Increment our counters:
            res[f"num_{ccs_type_stat_string}_reads"] += 1
            if read.has_tag(barcode_tag):
                res[f"num_{ccs_type_stat_string}_with_barcodes"] += 1
            if read.get_tag(longbow.utils.constants.COULD_CORRECT_BARCODE_TAG) == 1:
                res[f"num_{ccs_type_stat_string}_reads_corrected"] += 1
                if read.get_tag(longbow.utils.constants.BARCODE_CORRECTION_PERFORMED) == 0:
                    res[f"num_{ccs_type_stat_string}_reads_raw_was_correct"] += 1

            pbar.update(1)


def _correct_barcode_fn(in_queue, out_queue, bam_header_dict, barcode_tag, corrected_tag, barcode_length,
                        bc_allow_list, barcode_freqs, max_hifi_dist, max_clr_dist, ccs_corrected_rq_threshold):
    """Function to run in each subprocess.
    Replace barcode with corrected value."""

    bam_header = pysam.AlignmentHeader.from_dict(bam_header_dict)

    # Create the sym spell index here (in the sub-process) to reduce memory overhead / copying.
    logger.info(f"Generating barcode index...")
    st = time.time()
    sym_spell_index = barcode_utils.generate_symspell_index(barcode_freqs, max(max_hifi_dist, max_clr_dist),
                                                            barcode_length)
    logger.info(f"Barcode index generated in {time.time() - st:2.4f}s")

    while True:
        # Wait until we get some data.
        # Note: Because we have a sentinel value None inserted at the end of the input data for each
        #       subprocess, we don't have to add a timeout - we're guaranteed each process will always have
        #       at least one element.
        # try:
        #     raw_data = in_queue.get_nowait()
        # except queue.Empty:
        #     time.sleep(0.01)
        #     continue
        raw_data = in_queue.get()

        # Check for exit sentinel:
        if raw_data is None:
            return

        # Unpack our data here:
        read = pysam.AlignedSegment.fromstring(raw_data, bam_header)

        num_segments = 0
        num_corrected_segments = 0
        has_barcode = False

        if read.has_tag(barcode_tag):

            has_barcode = True
            old_bc = read.get_tag(barcode_tag)
            dist_threshold = max_hifi_dist if read.get_tag('rq') > ccs_corrected_rq_threshold else max_clr_dist

            offset = 0

            # Check to see if we have a barcode that is longer than expected.
            # If we do, then we should try perform matches at every valid position and take the best result:
            if len(old_bc) > barcode_length:
                new_bc, result_status, offset = _perform_barcode_multi_match(barcode_length, bc_allow_list,
                                                                             dist_threshold, old_bc, sym_spell_index)
            else:
                new_bc, _, result_status = \
                    barcode_utils.find_match_symspell(old_bc, bc_allow_list, sym_spell_index, dist_threshold)

            num_segments += 1
            if new_bc is not None:
                read.set_tag(corrected_tag, new_bc)
                read.set_tag(longbow.utils.constants.COULD_CORRECT_BARCODE_TAG, True)

                print(f"Old BC: {old_bc}")
                print(f"New BC: {new_bc}")

                # Need to do some basic math for padded barcode strings:
                if len(old_bc) > barcode_length:
                    end_pad_bases = int((len(old_bc) - barcode_length)/2)
                    correction_performed = new_bc != old_bc[end_pad_bases:end_pad_bases+barcode_length]
                else:
                    correction_performed = new_bc != old_bc
                read.set_tag(longbow.utils.constants.BARCODE_CORRECTION_PERFORMED, correction_performed)

                read.set_tag(longbow.utils.constants.READ_ADJUSTED_BARCODE_START, offset)

                num_corrected_segments += 1
            else:
                read.set_tag(longbow.utils.constants.COULD_CORRECT_BARCODE_TAG, False)
                read.set_tag(longbow.utils.constants.BARCODE_CORRECTION_PERFORMED, False)

                if result_status == barcode_utils.SymSpellMatchResultType.NO_MATCH_IN_LEV_DIST:
                    # NOTE: using the enum.value field here because of the TYPE of the barcode tag.
                    read.set_tag(corrected_tag, BarcodeResolutionFailure.NO_MATCH_IN_LEV_DIST.value)
                elif result_status == barcode_utils.SymSpellMatchResultType.AMBIGUOUS:
                    # NOTE: using the enum.value field here because of the TYPE of the barcode tag.
                    read.set_tag(corrected_tag, BarcodeResolutionFailure.AMBIGUOUS.value)
                else:
                    raise RuntimeError(f"Unknown SymSpellMatchResultType ({result_status})!  This should never happen!")
        else:
            read.set_tag(longbow.utils.constants.COULD_CORRECT_BARCODE_TAG, False)
            read.set_tag(longbow.utils.constants.BARCODE_CORRECTION_PERFORMED, False)
            # NOTE: using the enum.value field here because of the TYPE of the barcode tag.
            read.set_tag(corrected_tag, BarcodeResolutionFailure.NO_RAW_BARCODE.value)

        # Process and place our data on the output queue:
        out_queue.put(tuple([read.to_string(), num_segments, num_corrected_segments, has_barcode]))


def _perform_barcode_multi_match(barcode_length, bc_allow_list, dist_threshold, old_bc, sym_spell_index):
    # OK, we have to do multiple matches to try to find the barcode here:
    success_dict = dict()
    failure_counts = defaultdict(int)
    for i in range(0, len(old_bc) - barcode_length + 1):
        bc_sub_string = old_bc[i:i + barcode_length]

        new_bc, edit_dist, result_status = \
            barcode_utils.find_match_symspell(bc_sub_string, bc_allow_list, sym_spell_index, dist_threshold)
        if new_bc is not None:
            try:
                success_dict[edit_dist].append((new_bc, result_status, i))
            except KeyError:
                success_dict[edit_dist] = [(new_bc, result_status, i)]
        else:
            failure_counts[result_status] += 1

    # Setup for later:
    new_bc = None
    result_status = None
    offset = None

    # Now we can see if we have one good success:
    if len(success_dict) > 0:

        # We need to collapse entries with the same barcode here.
        # That is, it's possible that for different padded positions we're getting the same barcode back.
        # If this is the case, we should not call the correction ambiguous.

        lowest_dist = float("inf")
        collapsed_success_dict = dict()
        for lev_dist, match_info in success_dict.items():
            new_match_info = dict()

            # Arbitrarily keep only the first match:
            for barcode, status, index in match_info:
                if barcode not in new_match_info:
                    new_match_info[barcode] = (status, index)

            collapsed_success_dict[lev_dist] = [(barcode, status, index) for barcode, (status, index) in
                                                new_match_info.items()]

            # Since we're in here, let's just get the minimum distance now:
            if lev_dist < lowest_dist:
                lowest_dist = lev_dist

        # Now we can check that the lowest distance entry is a singleton:
        if len(collapsed_success_dict[lowest_dist]) == 1:
            new_bc = collapsed_success_dict[lowest_dist][0][0]
            result_status = collapsed_success_dict[lowest_dist][0][1]
            offset = collapsed_success_dict[lowest_dist][0][2]
        else:
            result_status = barcode_utils.SymSpellMatchResultType.AMBIGUOUS
    else:
        # Get the most prevalent failure type and report it:
        most_failures = -1
        for failure_type in failure_counts.keys():
            if failure_counts[failure_type] > most_failures:
                result_status = failure_type
                most_failures = failure_counts[failure_type]

    return new_bc, result_status, offset


def _get_barcode_tag_length_from_model(lb_model, barcode_tag):

    barcode_seg_name = lb_model.get_segment_name_for_annotation_tag(barcode_tag)
    if not barcode_seg_name:
        print(f"ERROR: Could not determine {lb_model.name} model segment from tag name: {barcode_tag}", file=sys.stderr)
        sys.exit(1)

    barcode_length = lb_model.get_segment_length(barcode_seg_name)
    if not barcode_length:
        print(f"ERROR: Could not extract {lb_model.name} model {barcode_seg_name} barcode segment from model.",
              file=sys.stderr)
        sys.exit(1)

    return barcode_length


def _generate_barcode_freqs_from_allow_list(barcode_allow_list):
    with tempfile.NamedTemporaryFile('w', prefix="tmp_barcode_allow_list_uniform_freqs",
                                     suffix=".tsv", dir=os.getcwd(), delete=False) as f:
        fname = f.name
        for bc in barcode_allow_list:
            f.write(f"{bc}\t1\n")

    return fname


def _filter_barcode_freqs_by_allow_list(barcode_freqs, barcode_allow_list):
    with open(barcode_freqs, 'r') as f:
        with tempfile.NamedTemporaryFile('w', prefix="tmp_barcode_freqs_filtered_to_allow_list",
                                         suffix=".tsv", dir=os.getcwd(), delete=False) as out_file:
            fname = out_file.name
            for line in f:
                bc = line.strip().split("\t")[0]
                if bc in barcode_allow_list:
                    out_file.write(line)
    return fname



import logging
import sys
import itertools
import time
import os
import math

import click
import click_log
import tqdm

import ssw

import pysam
import multiprocessing as mp

from ..utils import bam_utils
from ..utils.model import LibraryModel

import longbow.utils.constants

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("demultiplex")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(),
    help="BAM .pbi index file",
)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=mp.cpu_count() - 1,
    show_default=True,
    help="number of threads to use (0 for all)",
)
@click.option(
    "-o",
    "--out-base-name",
    default="longbow_demultiplexed",
    required=False,
    show_default=True,
    type=str,
    help="base name for output files",
)
@click.option(
    "-d",
    "--demux-on-tag",
    type=str,
    default="YN",
    show_default=True,
    required=False,
    help="BAM tag on which to demultiplex (e.g. YN, BC)"
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(pbi, threads, out_base_name, demux_on_tag, input_bam):
    """Separate reads into files based on which model they fit best.

    Resulting reads will be annotated with the model they best fit as well as the score and segments for that model."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    threads = mp.cpu_count() if threads <= 0 or threads > mp.cpu_count() else threads
    logger.info(f"Running with {threads} worker subprocess(es)")

    pbi = f"{input_bam.name}.pbi" if pbi is None else pbi
    read_count = None
    if os.path.exists(pbi):
        read_count = bam_utils.load_read_count(pbi)
        logger.info("Processing %d reads", read_count)

    # Create queues for data:
    queue_size = threads * 2 if threads < 10 else 20
    manager = mp.Manager()
    input_data_queue = manager.Queue(maxsize=queue_size)
    results = manager.Queue()

    # Start worker sub-processes:
    worker_pool = []

    for i in range(threads):
        p = mp.Process(
            target=_worker_demux_fn, args=(input_data_queue, results, i, demux_on_tag)
        )
        p.start()
        worker_pool.append(p)

    pysam.set_verbosity(0)  # silence message about the .bai file not being found
    with pysam.AlignmentFile(
        input_bam, "rb", check_sq=False, require_index=False
    ) as bam_file:

        # Get our header from the input bam file:
        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header)

        # Start output worker:
        res = manager.dict({"num_reads_demultiplexed": 0, "num_reads_discarded": 0})
        output_worker = mp.Process(
            target=_write_thread_fn, args=(results, out_header, out_base_name, res, read_count)
        )
        output_worker.start()

        # Add in a sentinel value at the end of the queue - one for each subprocess - so we guarantee
        # that all subprocesses will exit:
        iter_data = itertools.chain(bam_file, (None,) * threads)
        for r in iter_data:
            # We have to adjust for our sentinel value if we've got to it:
            if r is not None:
                r = r.to_string()
            input_data_queue.put(r)

        logger.debug("Finished reading data and sending it to sub-processes.")
        logger.debug("Waiting for sub-processes to finish...")

        # Wait for our input jobs to finish:
        for p in worker_pool:
            p.join()

        logger.debug("All workers stopped.")
        logger.debug("Terminating output process.")

        # Now that our input processes are done, we can add our exit sentinel onto the output queue and
        # wait for that process to end:
        results.put(None)
        output_worker.join()

    logger.info(
        f"Demultiplexed {res['num_reads_demultiplexed']} reads. Discarded {res['num_reads_discarded']}."
    )

    et = time.time()
    logger.info(f"Done. Elapsed time: {et - t_start:2.2f}s. "
                f"Overall processing rate: {(res['num_reads_demultiplexed']+res['num_reads_discarded'])/(et - t_start):2.2f} reads/s.")


def _write_thread_fn(out_queue, out_bam_header, out_bam_base_name, res, read_count):
    """Thread / process fn to write out all our data."""

    try:
        out_file_dict = {}

        with tqdm.tqdm(
            desc="Progress",
            unit=" read",
            colour="green",
            file=sys.stderr,
            total=read_count
        ) as pbar:

            while True:
                # Wait for some output data:
                raw_data = out_queue.get()

                # Check for exit sentinel:
                if raw_data is None:
                    break
                # Should really never be None, but just in case:
                elif raw_data is None:
                    continue

                # Unpack data:
                read, tag_value = raw_data

                if tag_value is not None:
                    read = pysam.AlignedSegment.fromstring(read, out_bam_header)

                    if tag_value not in out_file_dict:
                        out_file_dict[tag_value] = pysam.AlignmentFile(
                            f'{out_bam_base_name}.{tag_value}.bam',
                            "wb",
                            header=out_bam_header
                        )

                    out_file_dict[tag_value].write(read)

                    res["num_reads_demultiplexed"] += 1
                else:
                    res["num_reads_discarded"] += 1

                pbar.update(1)
    finally:
        for out_file in out_file_dict.values():
            out_file.close()


def _worker_demux_fn(in_queue, out_queue, worker_num, demux_on_tag):
    """Function to run in each subthread / subprocess.
    Simply pulls user-specified tag from read."""

    num_reads_processed = 0

    while True:
        # Wait until we get some data.
        # Note: Because we have a sentinel value None inserted at the end of the input data for each
        #       subprocess, we don't have to add a timeout - we're guaranteed each process will always have
        #       at least one element.
        raw_data = in_queue.get()

        # Check for exit sentinel:
        if raw_data is None:
            break
        # Should really never be None, but just in case:
        elif raw_data is None:
            continue

        # Unpack our data here:
        read = raw_data
        read = pysam.AlignedSegment.fromstring(
            read, pysam.AlignmentHeader.from_dict(dict())
        )

        tag_value = read.get_tag(demux_on_tag) if read.has_tag(demux_on_tag) else None

        out_queue.put((read.to_string(), tag_value))

        num_reads_processed += 1

    logger.debug(f"Worker %d: Num reads processed: %d", worker_num, num_reads_processed)



import pytest

import pysam
import tempfile
import os
import json

from longbow.utils import bam_utils
from longbow.utils import model


TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
)

@pytest.fixture
def bam_header_without_program_group():
    rgid = '01234567'
    movie_name = 'm00001e_210000_000000'

    header = pysam.AlignmentHeader.from_dict({
        'HD': {'VN': '1.5', 'SO': 'unknown', 'pb': '3.0.1'},
        'RG': [{
            'ID':rgid,
            'PL':'PACBIO',
            'DS':'READTYPE=CCS;Ipd:CodecV1=ip;PulseWidth:CodecV1=pw;BINDINGKIT=101-894-200;SEQUENCINGKIT=101-826-100;BASECALLERVERSION=5.0.0;FRAMERATEHZ=100.000000',
            'LB':'TestLib',
            'PU':movie_name,
            'SM':'TestSample',
            'PM':'SEQUELII',
            'CM':'S/P5-C2/5.0-8M'
        }],
        'PG': [{
            'ID':'ccs-6.0.0',
            'PN':'ccs',
            'VN':'6.0.0',
            'DS':'Generate circular consensus sequences (ccs) from subreads.',
            'CL':'ccs /opt/pacbio/pa-ccs/current/bin/ccs --all --streamed /data/pa/m00001e_210000_000000.consensusreadset.xml --bam /data/pa/m00001e_210000_000000.reads.bam --suppress-reports --num-threads 232 --log-level INFO --log-file /data/pa/m64020e_210000_000000.ccs.log --report-json /data/pa/m64020e_210000_000000.ccs_reports.json --report-file /data/pa/m64020e_210000_000000.ccs_reports.txt --metrics-json /data/pa/m64020e_210000_000000.zmw_metrics.json.gz --hifi-summary-json /data/pa/m64020e_210000_000000.hifi_summary.json --stderr-json-log --all-kinetics --subread-fallback'
        }],
        'SQ': []
    })

    return header


@pytest.fixture(scope="module", params=list(model.LibraryModel.pre_configured_models.keys()))
def bam_header_with_program_group(request):
    rgid = '01234567'
    movie_name = 'm00001e_210000_000000'
    version = '0.0.0'

    model_name = request.param
    model_json = model.LibraryModel.build_pre_configured_model(model_name).to_json(indent=None)

    header = pysam.AlignmentHeader.from_dict({
        'HD': {'VN': '1.5', 'SO': 'unknown', 'pb': '3.0.1'},
        'RG': [{
            'ID':rgid,
            'PL':'PACBIO',
            'DS':'READTYPE=CCS;Ipd:CodecV1=ip;PulseWidth:CodecV1=pw;BINDINGKIT=101-894-200;SEQUENCINGKIT=101-826-100;BASECALLERVERSION=5.0.0;FRAMERATEHZ=100.000000',
            'LB':'TestLib',
            'PU':movie_name,
            'SM':'TestSample',
            'PM':'SEQUELII',
            'CM':'S/P5-C2/5.0-8M'
        }],
        'PG': [{
            'ID':'ccs-6.0.0',
            'PN':'ccs',
            'VN':'6.0.0',
            'DS':'Generate circular consensus sequences (ccs) from subreads.',
            'CL':'ccs /opt/pacbio/pa-ccs/current/bin/ccs --all --streamed /data/pa/m00001e_210000_000000.consensusreadset.xml --bam /data/pa/m00001e_210000_000000.reads.bam --suppress-reports --num-threads 232 --log-level INFO --log-file /data/pa/m64020e_210000_000000.ccs.log --report-json /data/pa/m64020e_210000_000000.ccs_reports.json --report-file /data/pa/m64020e_210000_000000.ccs_reports.txt --metrics-json /data/pa/m64020e_210000_000000.zmw_metrics.json.gz --hifi-summary-json /data/pa/m64020e_210000_000000.hifi_summary.json --stderr-json-log --all-kinetics --subread-fallback'
        },
        {
            'ID':f'longbow-annotate-{version}',
            'PN':'longbow',
            'VN':version,
            'DS':f'Annotate reads in a BAM file with segments from the model.  MODEL(s): {model_json}',
            'CL':f'longbow annotate -m {model_name} -o test.ann.bam test.bam'
        }],
        'SQ': []
    })

    return header


@pytest.fixture(scope="module", params=list(filter(lambda x: x != "mas_15_sc_10x5p_single_none", model.LibraryModel.pre_configured_models.keys())))
def bam_header_with_multiple_program_groups(request):
    rgid = '01234567'
    movie_name = 'm00001e_210000_000000'
    version = '0.0.0'

    header_dict = {
        'HD': {'VN': '1.5', 'SO': 'unknown', 'pb': '3.0.1'},
        'RG': [{
            'ID':rgid,
            'PL':'PACBIO',
            'DS':'READTYPE=CCS;Ipd:CodecV1=ip;PulseWidth:CodecV1=pw;BINDINGKIT=101-894-200;SEQUENCINGKIT=101-826-100;BASECALLERVERSION=5.0.0;FRAMERATEHZ=100.000000',
            'LB':'TestLib',
            'PU':movie_name,
            'SM':'TestSample',
            'PM':'SEQUELII',
            'CM':'S/P5-C2/5.0-8M'
        }],
        'PG': [{
            'ID':'ccs-6.0.0',
            'PN':'ccs',
            'VN':'6.0.0',
            'DS':'Generate circular consensus sequences (ccs) from subreads.',
            'CL':'ccs /opt/pacbio/pa-ccs/current/bin/ccs --all --streamed /data/pa/m00001e_210000_000000.consensusreadset.xml --bam /data/pa/m00001e_210000_000000.reads.bam --suppress-reports --num-threads 232 --log-level INFO --log-file /data/pa/m64020e_210000_000000.ccs.log --report-json /data/pa/m64020e_210000_000000.ccs_reports.json --report-file /data/pa/m64020e_210000_000000.ccs_reports.txt --metrics-json /data/pa/m64020e_210000_000000.zmw_metrics.json.gz --hifi-summary-json /data/pa/m64020e_210000_000000.hifi_summary.json --stderr-json-log --all-kinetics --subread-fallback'
        }],
        'SQ': []
    }

    for model_name in ['mas_15_sc_10x5p_single_none', request.param]:
        model_json = model.LibraryModel.build_pre_configured_model(model_name).to_json(indent=None)

        header_dict['PG'].append({
            'ID':f'longbow-annotate-{version}',
            'PN':'longbow',
            'VN':version,
            'DS':f'Annotate reads in a BAM file with segments from the model.  MODEL(s): {model_json}',
            'CL':f'longbow annotate -m {model_name} -o test.ann.bam test.bam'
        })

    header = pysam.AlignmentHeader.from_dict(header_dict)

    return header


def test_load_read_count():
    pbi_file = 'tests/test_data/mas15_test_input.bam.pbi'
    n_reads_obs = bam_utils.load_read_count(pbi_file)
    n_reads_exp = 8

    assert n_reads_obs == n_reads_exp


def test_compute_shard_offsets():
    pbi_file = 'tests/test_data/mas15_test_input.bam.pbi'

    shard_offsets_exp = {
        2: [33947648, 921501696, 921547252],
        3: [33947648,  33995961, 921533666, 921547252],
        4: [33947648,  33980810, 921501696, 921533666, 921547252],
        5: [33947648,  33980810, 921501696, 921533666, 921547252]
    }

    for num_shards in range(2, 6):
        shard_offsets_obs, _, _, _, _ = bam_utils.compute_shard_offsets(pbi_file, num_shards)

        for i in range(len(shard_offsets_exp[num_shards])):
            assert shard_offsets_obs[i] == shard_offsets_exp[num_shards][i]


def test_bam_header_has_model(bam_header_with_program_group):
    ret = bam_utils.bam_header_has_model(bam_header_with_program_group)
    assert ret is True


def test_bam_header_is_missing_model(bam_header_without_program_group):
    ret = bam_utils.bam_header_has_model(bam_header_without_program_group)
    assert ret is False


def _compare_models(prebuilt_model, stored_model):
    assert prebuilt_model.name == stored_model.name
    assert prebuilt_model.description == stored_model.description
    assert prebuilt_model.version == stored_model.version
    assert prebuilt_model.array_element_structure == stored_model.array_element_structure
    assert prebuilt_model.direct_connections_dict == stored_model.direct_connections_dict
    assert prebuilt_model.end_element_names == stored_model.end_element_names
    assert prebuilt_model.key_adapter_set == stored_model.key_adapter_set
    assert prebuilt_model.key_adapters == stored_model.key_adapters
    assert prebuilt_model.start_element_names == stored_model.start_element_names

    assert prebuilt_model.adapter_dict.keys() == stored_model.adapter_dict.keys()
    for p, s in zip(prebuilt_model.adapter_dict.values(), stored_model.adapter_dict.values()):
        if isinstance(p, dict):
            assert p.keys() == s.keys()

            for k in p.keys():
                if isinstance(p[k], tuple):
                    # TODO: This is specifically to fix an issue comparing HomopolymerRepeat objects.
                    # Apparently, retrieving this from the pre-built model yields ('A', 30), but
                    # loading from json yields ['A', 30].  This should be fixed.
                    assert tuple(p[k]) == tuple(s[k])
                else:
                    assert p[k] == s[k]


@pytest.mark.parametrize("bam_header_with_multiple_program_groups", ['mas_15_bulk_teloprimeV2_single_none'], indirect=True)
def test_load_models_from_bam_header(bam_header_with_multiple_program_groups):
    with tempfile.NamedTemporaryFile(delete=True) as f:
        with pysam.AlignmentFile(f.name, "wb", header=bam_header_with_multiple_program_groups) as bf:
            bf.close()

        lb_models = bam_utils.load_models(None, input_bam=f)

        assert len(lb_models) == 2

        for lb_model in lb_models:
            stored_model = model.LibraryModel.from_json_file(f'{TEST_DATA_FOLDER}/models/{lb_model.name}.json')

            _compare_models(lb_model, stored_model)


def test_load_model_from_name():
    for model_name in list(model.LibraryModel.pre_configured_models.keys()):
        lb_models = bam_utils.load_models([model_name])

        stored_model = model.LibraryModel.from_json_file(f'{TEST_DATA_FOLDER}/models/{model_name}.json')

        _compare_models(lb_models[0], stored_model)


def test_load_model_from_json():
    json_string = b'{"name":"mas3teloprimev2","description":"An example model for a 3-element array.","version":"1.0.0","array_element_structure":[["A","TPV2_adapter","cDNA","Poly_A","idx","rev_bind"],["B","TPV2_adapter","cDNA","Poly_A","idx","rev_bind"], ["C","TPV2_adapter","cDNA","Poly_A","idx","rev_bind","D"]],"adapters":{"TPV2_adapter":"CTACACGACGCTCTTCCGATCTTGGATTGATATGTAATACGACTCACTATAG","rev_bind":"CTCTGCGTTGATACCACTGCTT","A":"AGCTTACTTGTGAAGAT","B":"ACTTGTAAGCTGTCTAT","C":"ACTCTGTCAGGTCCGAT","D":"ACCTCCTCCTCCAGAAT","Poly_A":{"HomopolymerRepeat":["A",30]},"idx":{"FixedLengthRandomBases":10},"cDNA":"random"},"direct_connections":{"A":["TPV2_adapter"],"B":["TPV2_adapter"],"C":["TPV2_adapter"],"D":["TPV2_adapter"],"TPV2_adapter":["cDNA"],"cDNA":["Poly_A"],"Poly_A":["idx"],"idx":["rev_bind"],"rev_bind":["B","C","D"]},"start_element_names":["A","TPV2_adapter"],"end_element_names":["rev_bind","D"],"named_random_segments":["idx","cDNA"],"coding_region":"cDNA","annotation_segments":{"idx":[["BC","XB"]]}}\n'
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(json_string)

    lb_models = bam_utils.load_models([f.name])
    stored_model = model.LibraryModel.from_json_obj(json.loads(json_string))

    _compare_models(lb_models[0], stored_model)


def test_reverse_complement():
    d = {
        'GTTCTAGCGCTAGTATG': 'CATACTAGCGCTAGAAC',
        'CTGCATAAAT': 'ATTTATGCAG',
        'TTTCGCGCATATAG': 'CTATATGCGCGAAA',
        'CCAA': 'TTGG',
        'TTATATTTAT': 'ATAAATATAA',
        'TANG': 'CNTA',
        'SGYACGR': 'YCGTRCS',
        'GCWATMCKVAABHGCD': 'HGCDVTTBMGKATWGC'
    }

    for k, v_exp in d.items():
        v_obs = bam_utils.reverse_complement(k)
        v_obs_lc = bam_utils.reverse_complement(k.lower())

        assert v_obs == v_exp
        assert v_obs_lc == v_exp.lower()


@pytest.mark.slow
def test_generate_read_name_produces_no_collisions():
    movie_name = 'm64020e_210000_000000'

    read_names = set()
    for zmw in range(8000000):
        for split_read_index in range(20):
            new_read_name = bam_utils.generate_read_name(movie_name, zmw, split_read_index)

            assert new_read_name not in read_names

            read_names.add(new_read_name)



import json
import logging
import time
import os
import sys

import click
import click_log

import tqdm
import pysam
from construct import *

import longbow.utils.constants
from ..utils import bam_utils
from ..utils import model as LongbowModel
from ..utils.model import LibraryModel
from ..annotate.command import get_segments

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("filter")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(exists=True),
    help="BAM .pbi index file",
)
@click.option(
    "-o",
    "--output-bam",
    default="-",
    type=click.Path(exists=False),
    help="filtered bam output (passing reads only)  [default: stdout]",
)
@click.option(
    "-x",
    "--reject-bam",
    default="/dev/null",
    type=click.Path(exists=False),
    help="Filtered bam output (failing reads only)  [default: /dev/null]",
)
@click.option(
    "-m",
    "--model",
    help="The model to use for annotation.  If not specified, it will be autodetected from "
         "the BAM header.  If the given value is a pre-configured model name, then that "
         "model will be used.  Otherwise, the given value will be treated as a file name "
         "and Longbow will attempt to read in the file and create a LibraryModel from it.  "
         "Longbow will assume the contents are the configuration of a LibraryModel as per "
         "LibraryModel.to_json()."
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(pbi, output_bam, reject_bam, model, force, input_bam):
    """Filter reads by conformation to expected segment order."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    pbi = f"{input_bam.name}.pbi" if pbi is None else pbi
    read_count = None
    if os.path.exists(pbi):
        read_count = bam_utils.load_read_count(pbi)
        logger.info("About to Filter %d reads", read_count)

    # Check to see if the output files exist:
    bam_utils.check_for_preexisting_files([output_bam, reject_bam], exist_ok=force)

    logger.info(f"Writing reads that conform to the model to: {output_bam}")
    logger.info(f"Writing reads that do not conform to the model to: {reject_bam}")

    # Open our input bam file:
    pysam.set_verbosity(0)
    with pysam.AlignmentFile(input_bam, "rb", check_sq=False, require_index=False) as bam_file, \
            tqdm.tqdm(
            desc="Progress",
            unit=" read",
            colour="green",
            file=sys.stderr,
            disable=not sys.stdin.isatty(),
            total=read_count
            ) as pbar:

        # Get our model:
        if model is None:
            lb_model = LibraryModel.from_json_obj(bam_utils.get_model_from_bam_header(bam_file.header))
        elif model is not None and LibraryModel.has_prebuilt_model(model):
            lb_model = LibraryModel.build_pre_configured_model(model)
        else:
            lb_model = LibraryModel.from_json_file(model)

        logger.info(f"Using %s: %s", lb_model.name, lb_model.description)

        logger.info(f"Filtering according to {lb_model.name} model ordered key adapters: "
                    f"{', '.join(lb_model.key_adapters)}")

        # Get our header from the input bam file:
        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header, models=[lb_model])

        # Setup output files:
        with pysam.AlignmentFile(output_bam, "wb", header=out_header) as passing_bam_file, \
                pysam.AlignmentFile(reject_bam, "wb", header=out_header) as failing_bam_file:

            num_passed = 0
            num_failed = 0

            tot_num_valid_adapters = 0
            tot_num_failed_adapters = 0

            for read in bam_file:
                # Get our read segments:
                try:
                    _, segments = get_segments(read)
                except KeyError:
                    logger.error(f"Input bam file does not contain longbow segmented reads!  "
                                 f"No {longbow.utils.constants.SEGMENTS_TAG} tag detected on read {read.query_name} !")
                    sys.exit(1)

                # Annotate the read with the model that was used in its validation:
                read.set_tag(longbow.utils.constants.READ_MODEL_NAME_TAG, lb_model.name)

                # Check to see if the read is valid by this model and write it out:
                segment_names = [s.name for s in segments]
                is_valid, num_valid_adapters, first_valid_adapter_index = \
                    lb_model.validate_segment_order(segment_names)

                if is_valid:
                    logger.debug("Read is %s valid: %s: first key adapter: [%d, %s], # key adapters: %d",
                                 lb_model.name,
                                 read.query_name,
                                 first_valid_adapter_index,
                                 lb_model.key_adapters[first_valid_adapter_index],
                                 num_valid_adapters)

                    read.set_tag(longbow.utils.constants.READ_IS_VALID_FOR_MODEL_TAG, True)
                    read.set_tag(longbow.utils.constants.READ_NUM_KEY_SEGMENTS_TAG, num_valid_adapters)
                    read.set_tag(longbow.utils.constants.READ_FIRST_KEY_SEG_TAG, lb_model.key_adapters[first_valid_adapter_index])
                    passing_bam_file.write(read)
                    tot_num_valid_adapters += num_valid_adapters
                    num_passed += 1
                else:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Read is not %s valid: %s: first key adapter: [%d, %s], # key adapters: %d, "
                                     "key adapters detected: %s",
                                     lb_model.name,
                                     read.query_name,
                                     first_valid_adapter_index,
                                     lb_model.key_adapters[first_valid_adapter_index],
                                     num_valid_adapters,
                                     lb_model.extract_key_segment_names(segment_names))

                    read.set_tag(longbow.utils.constants.READ_IS_VALID_FOR_MODEL_TAG, False)
                    failing_bam_file.write(read)
                    tot_num_failed_adapters += num_valid_adapters
                    num_failed += 1

                pbar.update(1)

    # Calc some stats:
    pct_reads_passing = 100 * num_passed / (num_passed + num_failed) if (num_passed + num_failed) > 0 else 0
    pct_reads_failing = 100 * num_failed / (num_passed + num_failed) if (num_passed + num_failed) > 0 else 0
    mean_adapters_per_passing_read = tot_num_valid_adapters/num_passed if num_passed > 0 else 0
    mean_adapters_per_failing_read = tot_num_failed_adapters/num_failed if num_failed > 0 else 0

    # Yell at the user:
    logger.info(f"Done. Elapsed time: %2.2fs.", time.time() - t_start)
    logger.info(f"Total Reads Processed: %d", num_passed + num_failed)
    logger.info(f"# Reads Passing Model Filter: %d (%2.2f%%)", num_passed, pct_reads_passing)
    logger.info(f"# Reads Failing Model Filter: %d (%2.2f%%)", num_failed, pct_reads_failing)
    logger.info(f"Total # correctly ordered key adapters in passing reads: %d", tot_num_valid_adapters)
    logger.info(f"Total # correctly ordered key adapters in failing reads: %d", tot_num_failed_adapters)
    logger.info(f"Avg # correctly ordered key adapters per passing read: %2.4f [%d]",
                mean_adapters_per_passing_read, len(lb_model.key_adapters))
    logger.info(f"Avg # correctly ordered key adapters per failing read: %2.4f [%d]",
                mean_adapters_per_failing_read, len(lb_model.key_adapters))



import pytest
import os
import sys
import subprocess
import tempfile

from click.testing import CliRunner

from longbow.__main__ import main_entry as longbow

from ..utils import cat_file_to_pipe
from ..utils import convert_sam_to_bam
from ..utils import assert_reads_files_equal

################################################################################

TOOL_NAME = "segment"

TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
) + os.path.sep
EXPECTED_DATA_FOLDER = TEST_DATA_FOLDER + TOOL_NAME + os.path.sep

################################################################################


@pytest.fixture(scope="module", params=[
    (TEST_DATA_FOLDER + "mas15_test_input.bam",
     EXPECTED_DATA_FOLDER + "mas_15_sc_10x5p_single_none.expected.bam",
     "mas_15_sc_10x5p_single_none"),

    (TEST_DATA_FOLDER + "mas10_test_input.bam",
     EXPECTED_DATA_FOLDER + "mas_10_sc_10x5p_single_none.expected.bam",
     "mas_10_sc_10x5p_single_none"),

    (TEST_DATA_FOLDER + "mas_15_bulk_10x5p_single_internal.bam",
     EXPECTED_DATA_FOLDER + "mas_15_bulk_10x5p_single_internal.expected.bam",
     "mas_15_bulk_10x5p_single_internal"),
])
def filtered_bam_file_from_pipeline(request):
    input_file, expected_file, model_name = request.param

    if input_file.endswith(".bam"):
        input_bam = input_file
    else:
        # Convert test file to bam:
        with tempfile.NamedTemporaryFile(delete=True) as input_bam:
            convert_sam_to_bam(input_file, input_bam)

    with tempfile.NamedTemporaryFile(delete=True) as annotate_bam, \
         tempfile.NamedTemporaryFile(delete=True) as filter_bam:

        runner = CliRunner()

        result_annotate = runner.invoke(longbow, ["annotate", "-t", 1, "-m", model_name,
                                                  "-f", "-o", annotate_bam.name, input_bam])
        assert result_annotate.exit_code == 0

        result_filter = runner.invoke(longbow, ["filter", "-m", model_name,
                                                "-f", "-o", filter_bam.name, annotate_bam.name])
        assert result_filter.exit_code == 0

        # Yield file here so that when we return, we get to clean up automatically
        yield filter_bam.name, expected_file, model_name


def test_segment_from_file(tmpdir, filtered_bam_file_from_pipeline):
    input_bam_file, expected_bam, model_name = filtered_bam_file_from_pipeline

    actual_file = tmpdir.join(f"segment_actual_out.bam")
    args = ["segment", "-t", 1, "-f", "-o", actual_file, input_bam_file]

    runner = CliRunner()
    result = runner.invoke(longbow, args)

    assert result.exit_code == 0
    assert_reads_files_equal(actual_file, expected_bam, order_matters=True)


def test_segment_from_pipe(tmpdir, filtered_bam_file_from_pipeline):
    input_bam_file, expected_bam, model_name = filtered_bam_file_from_pipeline

    actual_file = tmpdir.join(f"filter_actual_out.pipe.bam")

    proc = subprocess.Popen(
        [sys.executable, "-m", "longbow", "segment", "-t", "1", "-f", "-o", actual_file],
        stdin=subprocess.PIPE
    )

    cat_file_to_pipe(input_bam_file, proc)

    assert proc.returncode == 0
    assert_reads_files_equal(actual_file, expected_bam, order_matters=True)
